/*
 *  X86 code generator for TCC
 * 
 *  Copyright (c) 2001-2004 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef TARGET_DEFS_ONLY

/* number of available registers */
#define NB_REGS         5
#define NB_ASM_REGS     8
#define CONFIG_TCC_ASM

/* a register can belong to several classes. The classes must be
   sorted from more general to more precise (see gv2() code which does
   assumptions on it). */
#define RC_INT     0x0001 /* generic integer register */
#define RC_FLOAT   0x0002 /* generic float register */
#define RC_EAX     0x0004
#define RC_ST0     0x0008 
#define RC_ECX     0x0010
#define RC_EDX     0x0020
#define RC_EBX     0x0040

#define RC_IRET    RC_EAX /* function return: integer register */
#define RC_LRET    RC_EDX /* function return: second integer register */
#define RC_FRET    RC_ST0 /* function return: float register */

/* pretty names for the registers */
enum {
    TREG_EAX = 0,
    TREG_ECX,
    TREG_EDX,
    TREG_EBX,
    TREG_ST0,
    TREG_ESP = 4
};

/* return registers for function */
#define REG_IRET TREG_EAX /* single word int return register */
#define REG_LRET TREG_EDX /* second word return register (for long long) */
#define REG_FRET TREG_ST0 /* float return register */

/* defined if function parameters must be evaluated in reverse order */
#define INVERT_FUNC_PARAMS

/* defined if structures are passed as pointers. Otherwise structures
   are directly pushed on stack. */
/* #define FUNC_STRUCT_PARAM_AS_PTR */

/* pointer size, in bytes */
#define PTR_SIZE 4

/* long double size and alignment, in bytes */
#define LDOUBLE_SIZE  12
#define LDOUBLE_ALIGN 4
/* maximum alignment (for aligned attribute support) */
#define MAX_ALIGN     8

/******************************************************/
#else /* ! TARGET_DEFS_ONLY */
/******************************************************/
#include "tcc.h"

/* define to 1/0 to [not] have EBX as 4th register */
#define USE_EBX 0

ST_DATA const int reg_classes[NB_REGS] = {
    /* eax */ RC_INT | RC_EAX,
    /* ecx */ RC_INT | RC_ECX,
    /* edx */ RC_INT | RC_EDX,
    /* ebx */ (RC_INT | RC_EBX) * USE_EBX,
    /* st0 */ RC_FLOAT | RC_ST0,
};

ST_DATA prolog_t prolog;
static unsigned long func_sub_sp_offset;
static int func_ret_sub;
#ifdef CONFIG_TCC_BCHECK
static addr_t func_bound_offset;
static unsigned long func_bound_ind;
#endif

/* XXX: make it faster ? */
ST_FUNC void g(int c)
{
    int ind1;
    if (nocode_wanted)
        return;
    ind1 = ind + 1;
    if (ind1 > cur_text_section->data_allocated)
        section_realloc(cur_text_section, ind1);
    cur_text_section->data[ind] = c;
    ind = ind1;
}

ST_FUNC void o(unsigned int c)
{
    while (c) {
        g(c);
        c = c >> 8;
    }
}

ST_FUNC void gen_le16(int v)
{
    g(v);
    g(v >> 8);
}

ST_FUNC void gen_le32(int c)
{
    g(c);
    g(c >> 8);
    g(c >> 16);
    g(c >> 24);
}

/* output a symbol and patch all calls to it */
ST_FUNC void gsym_addr(int t, int a)
{
    while (t) {
        unsigned char *ptr = cur_text_section->data + t;
        uint32_t n = read32le(ptr); /* next value */
        write32le(ptr, a - t - 4);
        t = n;
    }
}

ST_FUNC void gsym(int t)
{
    gsym_addr(t, ind);
}

/* instruction + 4 bytes data. Return the address of the data */
static int oad(int c, int s)
{
    int t;
    if (nocode_wanted)
        return s;
    o(c);
    t = ind;
    gen_le32(s);
    return t;
}

/* generate jmp to a label */
#define gjmp2(instr,lbl) oad(instr,lbl)

/* output constant with relocation if 'r & VT_SYM' is true */
ST_FUNC void gen_addr32(int r, Sym *sym, int c)
{
    if (r & VT_SYM)
        greloc(cur_text_section, sym, ind, R_386_32);
    gen_le32(c);
}

ST_FUNC void gen_addrpc32(int r, Sym *sym, int c)
{
    if (r & VT_SYM)
        greloc(cur_text_section, sym, ind, R_386_PC32);
    gen_le32(c - 4);
}

/* generate a modrm reference. 'op_reg' contains the additional 3
   opcode bits */
static void gen_modrm(int op_reg, int r, Sym *sym, int c)
{
    op_reg = op_reg << 3;
    if ((r & VT_VALMASK) == VT_CONST) {
        /* constant memory reference */
	if( xx.sw.para )
	    o(0x83 | op_reg);
	else
        o(0x05 | op_reg);
        gen_addr32(r, sym, c);
    } else if ((r & VT_VALMASK) == VT_LOCAL) {
	if( xx.sw.prolog )
	    c -= prolog.ebp_offset;
        /* currently, we use only ebp as base */
        if (c == (char)c) {
            /* short reference */
            o(0x45 | op_reg);
            g(c);
        } else {
            oad(0x85 | op_reg, c);
        }
    } else {
        g(0x00 | op_reg | (r & VT_VALMASK));
    }
}

#define exchange_reg(x, y) { tmp = R[x]; R[x] = R[y]; R[y] = tmp; }

ST_FUNC void add_descrambler(int r )
{
  u32 rnd[3], tmp, m, i, j, R[8];
  u32 r0sh, r1sh, r2sh, r1r0sh, r1r1sh, r2r0sh;

  read( rnd_fd, rnd, 3 * sizeof( u32 ) );
  // eax, ecx, edx, ebx, esp, ebp, esi, edi
  for( i = 0; i < 8; i++ )
    R[i] = i;

  exchange_reg( 0, r );	// parameter in R0
  m = 1;
  for( i = 1; i < 8; i++ ) // shuffle the other registers
  {
    if( i == 4 ) // don't use esp
      continue;
    for( j = i + 1; j < 8; j++ )
    {
      if( j == 4 ) // don't use esp
	continue;
      if( rnd[0] & m )
	exchange_reg( i, j );
      m <<= 1;
    }
  }

  r1r0sh = R[1]<<11 | R[0]<<8;
  r1r1sh = R[1]<<19 | R[1]<<16;
  r2r0sh = R[2]<<11 | R[0]<<8;
  r0sh = R[0]<<8;
  r1sh = R[1]<<8;
  r2sh = R[2]<<8;

  o( 0x50 | R[1] );		//  push  ecx
  oad( 0xb8 | R[1], rnd[1] );	//  mov	  ecx, 0xae4c865d
  oad( 0xf081 | r1sh, rnd[1] ^ 0xae4c865d );  //  xor	ecx, 0xae4c865d
  o( 0x0033 | r1r0sh );		//  xor	  ecx, [ebx]
  o( 0x044033 | r1r0sh );	//  xor	  ecx, [ebx+4]
  o( 0xe8d1 | r1sh );		//  shr	  ecx, 1
  o( 0x3e75 );			//  jnz	  normal
  o( 0x0573 );			//  jnc	  test_len
  o( 0x10c083 | r0sh );		//  add	  ebx, 0x10
  o( 0x37eb );			//  jmp	  normal
				// test_len:
  o( 0x008b | r1r0sh );		//  mov	  ecx, [ebx]
  o( 0xe8d1 | r1sh );		//  shr	  ecx, 1
  o( 0x084033 | r1r0sh );	//  xor	  ecx, [ebx+8]
  o( 0x10c0c1 | r1sh );		//  rol	  ecx, 16
  o( 0xc08566 | r1r1sh );	//  test  cx, cx
  o( 0x2875 );			//  jnz	  normal
				// decrypt:
  o( 0x10c0c1 | r1sh );		//  rol	  ecx, 16 ; length in bytes
  o( 0x010883 | r0sh );		//  or	  dword [ebx], 1 ; set lsb
  o( 0x10c083 | r0sh );		//  add	  ebx, 0x10
  o( 0x50 | R[0] );		//  push  ebx
  o( 0x50 | R[2] );		//  push  edx
  o( 0xfc408b | r2r0sh );	//  mov	  edx, [ebx-4]	; lfsr seed
				// next_word:
  o( 0x0031 | r2r0sh );		//  xor	  [ebx], edx
  o( 0x04c083 | r0sh );		//  add	  ebx, 4
  o( 0xe8d1 | r2sh );		//  shr	  edx, 1 ; lfsr
  o( 0x0c73 );			//  jnc	  no_taps
  oad( 0xf081 | r2sh, rnd[2] ); //  xor	  edx, 0xa451c123
  oad( 0xf081 | r2sh, rnd[2] ^ 0xa451c123 ); //	 xor   edx, 0xa451c123
				// no_taps:
  o( 0x48 | R[1] );		//  dec	  ecx
  o( 0xe875 );			//  jnz	  next_word
  o( 0x58 | R[2] );		//  pop	  edx
  o( 0x58 | R[0] );		//  pop	  ebx
				// normal:
  o( 0x58 | R[1] );		//  pop	  ecx
}

/* load 'r' from value 'sv' */
ST_FUNC void load(int r, SValue *sv)
{
    int v, t, ft, fc, fr;
    SValue v1;
    u32 rnd[2];

#ifdef TCC_TARGET_PE
    SValue v2;
    sv = pe_getimport(sv, &v2);
#endif

    if( xx.sw.para )
       read( rnd_fd, rnd, 2 * sizeof( u32 ) );
    fr = sv->r;
    ft = sv->type.t & ~VT_DEFSIGN;
    fc = sv->c.i;

    ft &= ~(VT_VOLATILE | VT_CONSTANT);

    v = fr & VT_VALMASK;
    if (fr & VT_LVAL) {
        if (v == VT_LLOCAL) {
            v1.type.t = VT_INT;
            v1.r = VT_LOCAL | VT_LVAL;
            v1.c.i = fc;
            fr = r;
            if (!(reg_classes[fr] & RC_INT))
                fr = get_reg(RC_INT);
            load(fr, &v1);
        }
	if( xx.sw.para ) {
	    if (v == VT_CONST) {
		oad( 0xbb, rnd[0] );
		fc -= rnd[0];
	    }
	}
        if ((ft & VT_BTYPE) == VT_FLOAT) {
            o(0xd9); /* flds */
            r = 0;
        } else if ((ft & VT_BTYPE) == VT_DOUBLE) {
            o(0xdd); /* fldl */
            r = 0;
        } else if ((ft & VT_BTYPE) == VT_LDOUBLE) {
            o(0xdb); /* fldt */
            r = 5;
        } else if ((ft & VT_TYPE) == VT_BYTE || (ft & VT_TYPE) == VT_BOOL) {
            o(0xbe0f);   /* movsbl */
        } else if ((ft & VT_TYPE) == (VT_BYTE | VT_UNSIGNED)) {
            o(0xb60f);   /* movzbl */
        } else if ((ft & VT_TYPE) == VT_SHORT) {
            o(0xbf0f);   /* movswl */
        } else if ((ft & VT_TYPE) == (VT_SHORT | VT_UNSIGNED)) {
            o(0xb70f);   /* movzwl */
        } else {
            o(0x8b);     /* movl */
        }
        gen_modrm(r, fr, sv->sym, fc);
    } else {
        if (v == VT_CONST) {
	    if( xx.sw.para ) {
		oad(0xbb, rnd[0] ); // mov ebx, $x
		o(0xd887 | r<<8 );  // xchg ebx, r
		oad(0xbb, rnd[1] ); // mov ebx, $x
		o(0xd887 | r<<8 ); // xchg ebx, r
		o(0xd831 | r<<8 ); // xor ebx, r
		o(0xd887 | r<<8 ); // xchg ebx, r
		o(0xc381 );	   // add ebx,
		gen_addr32(fr, sv->sym, fc - (rnd[0] ^ rnd[1]) );
		o(0xd887 | r<<8 ); // xchg ebx, r
	    } else {
            o(0xb8 + r); /* mov $xx, r */
            gen_addr32(fr, sv->sym, fc);
	    }
        } else if (v == VT_LOCAL) {
            if (fc) {
                o(0x8d); /* lea xxx(%ebp), r */
                gen_modrm(r, VT_LOCAL, sv->sym, fc);
            } else {
                o(0x89);
                o(0xe8 + r); /* mov %ebp, r */
            }
        } else if (v == VT_CMP) {
            oad(0xb8 + r, 0); /* mov $0, r */
            o(0x0f); /* setxx %br */
            o(fc);
            o(0xc0 + r);
        } else if (v == VT_JMP || v == VT_JMPI) {
            t = v & 1;
            oad(0xb8 + r, t); /* mov $1, r */
            o(0x05eb); /* jmp after */
            gsym(fc);
            oad(0xb8 + r, t ^ 1); /* mov $0, r */
        } else if (v != r) {
            o(0x89);
            o(0xc0 + r + v * 8); /* mov v, r */
        }
    }
    ft = sv->type.t & VT_BTYPE;
    if( xx.sw.lfsr && (ft == VT_PTR || ft == VT_LLONG ) )
    {
      if( sv->sym && (sv->sym->type.t & VT_BTYPE) == VT_PTR ) {
        add_descrambler( r );
      }
    }
}

/* store register 'r' in lvalue 'v' */
ST_FUNC void store(int r, SValue *v)
{
    int fr, bt, ft, fc;
    u32 rnd;

#ifdef TCC_TARGET_PE
    SValue v2;
    v = pe_getimport(v, &v2);
#endif

    if( xx.sw.para )
      read( rnd_fd, &rnd, sizeof( u32 ) );
    ft = v->type.t;
    fc = v->c.i;
    fr = v->r & VT_VALMASK;
    if( xx.sw.para ) {
      if (fr == VT_CONST) {
	oad( 0xbb, rnd );
	fc -= rnd;
      }
    }
    ft &= ~(VT_VOLATILE | VT_CONSTANT);
    bt = ft & VT_BTYPE;
    /* XXX: incorrect if float reg to reg */
    if (bt == VT_FLOAT) {
        o(0xd9); /* fsts */
        r = 2;
    } else if (bt == VT_DOUBLE) {
        o(0xdd); /* fstpl */
        r = 2;
    } else if (bt == VT_LDOUBLE) {
        o(0xc0d9); /* fld %st(0) */
        o(0xdb); /* fstpt */
        r = 7;
    } else {
        if (bt == VT_SHORT)
            o(0x66);
        if (bt == VT_BYTE || bt == VT_BOOL)
            o(0x88);
        else
            o(0x89);
    }
    if (fr == VT_CONST ||
        fr == VT_LOCAL ||
        (v->r & VT_LVAL)) {
        gen_modrm(r, v->r, v->sym, fc);
    } else if (fr != r) {
        o(0xc0 + fr + r * 8); /* mov r, fr */
    }
}

static void gadd_sp(int val)
{
  u8 rnd[0x28];
  int i;
  if( xx.sw.add_sp ) {
    read( rnd_fd, rnd, 0x28 );
    rnd[4] &= 0xf;
    rnd[5] &= 0xf;
    oad(0xec81, *(u32 *)rnd ); /* sub $xxx, %esp */
    o(0x75);	    /* jnz */
    o( rnd[4] + 1 );
    for( i = 0; i <= rnd[4]; i++ )
      g( rnd[0x08+i] );
    oad(0xc481, *(u32 *)rnd + val ); /* add $xxx, %esp */
    o(0x75);	    /* jnz */
    o( rnd[5] + 1 );
    for( i = 0; i <= rnd[5]; i++ )
      g( rnd[0x18+i] );
  } else {
    if (val == (char)val) {
        o(0xc483);
        g(val);
    } else {
        oad(0xc481, val); /* add $xxx, %esp */
    }
  }
}

#if defined CONFIG_TCC_BCHECK || defined TCC_TARGET_PE
static void gen_static_call(int v)
{
    Sym *sym;

    sym = external_global_sym(v, &func_old_type, 0);
    oad(0xe8, -4);
    greloc(cur_text_section, sym, ind-4, R_386_PC32);
}
#endif

/* 'is_jmp' is '1' if it is a jump */
static void gcall_or_jmp(int is_jmp)
{
    int i, r;
    u8 rnd[0x48];

    if ((vtop->r & (VT_VALMASK | VT_LVAL)) == VT_CONST && (vtop->r & VT_SYM)) {
	if( xx.sw.call ) {
	   read( rnd_fd, rnd, 0x48 );
	   rnd[4] &= 0xf;
	   rnd[5] &= 0xf;
	   rnd[6] &= 0xf;
	   rnd[7] &= 0xf;
	   oad(0xe8, rnd[4] + 1 );
	   for( i = 0; i <= rnd[4]; i++ )
	     g( rnd[0x08+i] );
	   o(0x5b);	       // pop ebx
	   o(0xdb21);	       // and ebx, ebx
	   o(0x75);	       // jnz
	   o( rnd[5] + 1 );
	   for( i = 0; i <= rnd[5]; i++ )
	     g( rnd[0x18+i] );
	   o(0x81);	       /* add ebx, */
	}
        /* constant and relocation case */
        greloc(cur_text_section, vtop->sym, ind + 1, R_386_PC32);
	if( xx.sw.call ) {
	   oad(0xc3, *(u32 *)rnd + vtop->c.i + rnd[4] + rnd[5] + 9);
	   o(0x75);			       /* jnz */
	   o( rnd[6] + 1 );
	   for( i = 0; i <= rnd[6]; i++ )
	     g( rnd[0x28+i] );
	   oad(0xeb81, *(u32 *)rnd);	  /* sub ebx, */
	   o(0x75);			       /* jnz */
	   o( rnd[7] + 1 );
	   for( i = 0; i <= rnd[7]; i++ )
	     g( rnd[0x38+i] );
	   o(0xff);
	   o(0xd3 + (is_jmp << 4));
	} else {
        oad(0xe8 + is_jmp, vtop->c.i - 4); /* call/jmp im */
	}
    } else {
        /* otherwise, indirect call */
        r = gv(RC_INT);
        o(0xff); /* call/jmp *r */
        o(0xd0 + r + (is_jmp << 4));
    }
    if (!is_jmp) {
        int rt;
        /* extend the return value to the whole register if necessary
           visual studio and gcc do not always set the whole eax register
           when assigning the return value of a function  */
        rt = vtop->type.ref->type.t;
        switch (rt & VT_BTYPE) {
            case VT_BYTE:
                if (rt & VT_UNSIGNED) {
                    o(0xc0b60f); /* movzx %al, %eax */
                }
                else {
                    o(0xc0be0f); /* movsx %al, %eax */
                }
                break;
            case VT_SHORT:
                if (rt & VT_UNSIGNED) {
                    o(0xc0b70f); /* movzx %ax, %eax */
                }
                else {
                    o(0xc0bf0f); /* movsx %ax, %eax */
                }
                break;
            default:
                break;
        }
    }
}

static uint8_t fastcall_regs[3] = { TREG_EAX, TREG_EDX, TREG_ECX };
static uint8_t fastcallw_regs[2] = { TREG_ECX, TREG_EDX };

/* Return the number of registers needed to return the struct, or 0 if
   returning via struct pointer. */
ST_FUNC int gfunc_sret(CType *vt, int variadic, CType *ret, int *ret_align, int *regsize)
{
#ifdef TCC_TARGET_PE
    int size, align;
    *ret_align = 1; // Never have to re-align return values for x86
    *regsize = 4;
    size = type_size(vt, &align);
    if (size > 8 || (size & (size - 1)))
        return 0;
    if (size == 8)
        ret->t = VT_LLONG;
    else if (size == 4)
        ret->t = VT_INT;
    else if (size == 2)
        ret->t = VT_SHORT;
    else
        ret->t = VT_BYTE;
    ret->ref = NULL;
    return 1;
#else
    *ret_align = 1; // Never have to re-align return values for x86
    return 0;
#endif
}

/* Generate function call. The function address is pushed first, then
   all the parameters in call order. This functions pops all the
   parameters and the function address. */
ST_FUNC void gfunc_call(int nb_args)
{
    int size, align, r, args_size, i, j, func_call;
    Sym *func_sym;
    u8 rnd[0x18];
    
    args_size = 0;
    for(i = 0;i < nb_args; i++) {
	if( xx.sw.func )
	    read( rnd_fd, rnd, 0x18 );
        if ((vtop->type.t & VT_BTYPE) == VT_STRUCT) {
            size = type_size(&vtop->type, &align);
            /* align to stack align size */
            size = (size + 3) & ~3;
            /* allocate the necessary size on stack */
	    if( xx.sw.func ) {
		rnd[4] &= 0xf;
		oad(0xc481, *(u32 *)rnd );
		o(0x75);		/* jnz */
		o( rnd[4] + 1 );
		for( j = 0; j <= rnd[4]; j++ )
		  g( rnd[0x08+j] );
		oad(0xec81, *(u32 *)rnd + size );
	    } else
            oad(0xec81, size); /* sub $xxx, %esp */
            /* generate structure store */
            r = get_reg(RC_INT);
            o(0x89); /* mov %esp, r */
            o(0xe0 + r);
            vset(&vtop->type, r | VT_LVAL, 0);
            vswap();
            vstore();
            args_size += size;
        } else if (is_float(vtop->type.t)) {
            gv(RC_FLOAT); /* only one float register */
            if ((vtop->type.t & VT_BTYPE) == VT_FLOAT)
                size = 4;
            else if ((vtop->type.t & VT_BTYPE) == VT_DOUBLE)
                size = 8;
            else
                size = 12;
	    if( xx.sw.func ) {
		rnd[4] &= 0xf;
		oad(0xc481, *(u32 *)rnd );
		o(0x75);    /* jnz */
		o( rnd[4] + 1 );
		for( j = 0; j <= rnd[4]; j++ )
		  g( rnd[0x08+j] );
		oad(0xec81, *(u32 *)rnd + size );
	    } else
            oad(0xec81, size); /* sub $xxx, %esp */
            if (size == 12)
                o(0x7cdb);
            else
                o(0x5cd9 + size - 4); /* fstp[s|l] 0(%esp) */
            g(0x24);
            g(0x00);
            args_size += size;
        } else {
            /* simple type (currently always same size) */
            /* XXX: implicit cast ? */
            r = gv(RC_INT);
            if ((vtop->type.t & VT_BTYPE) == VT_LLONG) {
                size = 8;
                o(0x50 + vtop->r2); /* push r */
            } else {
                size = 4;
            }
            o(0x50 + r); /* push r */
            args_size += size;
        }
        vtop--;
    }
    save_regs(0); /* save used temporary registers */
    func_sym = vtop->type.ref;
    func_call = func_sym->f.func_call;
    /* fast call case */
    if ((func_call >= FUNC_FASTCALL1 && func_call <= FUNC_FASTCALL3) ||
        func_call == FUNC_FASTCALLW) {
        int fastcall_nb_regs;
        uint8_t *fastcall_regs_ptr;
        if (func_call == FUNC_FASTCALLW) {
            fastcall_regs_ptr = fastcallw_regs;
            fastcall_nb_regs = 2;
        } else {
            fastcall_regs_ptr = fastcall_regs;
            fastcall_nb_regs = func_call - FUNC_FASTCALL1 + 1;
        }
        for(i = 0;i < fastcall_nb_regs; i++) {
            if (args_size <= 0)
                break;
            o(0x58 + fastcall_regs_ptr[i]); /* pop r */
            /* XXX: incorrect for struct/floats */
            args_size -= 4;
        }
    }
#ifndef TCC_TARGET_PE
    else if ((vtop->type.ref->type.t & VT_BTYPE) == VT_STRUCT)
        args_size -= 4;
#endif
    gcall_or_jmp(0);

    if (args_size && func_call != FUNC_STDCALL && func_call != FUNC_FASTCALLW)
        gadd_sp(args_size);
    vtop--;
}

#ifdef TCC_TARGET_PE
#define FUNC_PROLOG_SIZE 11
#define X_PROLOG_SIZE 28
#else
#define FUNC_PROLOG_SIZE 10
#define X_PROLOG_SIZE 22
#endif

void emit_unscramble_data( CType *func_type )
{
#define XDATA_RND_BUF (5 * 4 + 9 * 17 )
    u8 rnd[XDATA_RND_BUF];
    u32 *wrnd;
    u8 *brnd, *irnd;
    Sym *sym;
    int i;

    wrnd = (u32 *)rnd;
    brnd = irnd = rnd + 0x14;
    read( rnd_fd, rnd, XDATA_RND_BUF );
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    *irnd++ &= 0xf;
    prolog.code_start = ind;
    sym = get_sym_ref(func_type, data_section, 0, 1);	// dummy symbol
    oad(0xbb, wrnd[0] );	// mov ebx
    put_elf_reloc(symtab_section, cur_text_section, ind-4, R_386_32, sym->c );
    o(0x75db85);		// test ebx, ebx, jnz
    g( brnd[0] + 1 );
    for( i = 0; i <= brnd[0]; i++ )
       g( *irnd++ );
    oad(0xc381, -wrnd[0] );	// add	ebx
    g(0x75);			// jnz
    g( brnd[1] + 1 );
    for( i = 0; i <= brnd[1]; i++ )
       g( *irnd++ );
    oad(0xb9, wrnd[1] );	// mov ecx
    prolog.length_entry = ind - 4;
    o(0x75c985);		// test ecx, ecx, jnz
    g( brnd[2] + 1 );
    for( i = 0; i <= brnd[2]; i++ )
       g( *irnd++ );
    oad(0xc181, -wrnd[1] );	// add ecx
    o(0x75);			// jnz
    g( brnd[3] + 1 );
    for( i = 0; i <= brnd[3]; i++ )
       g( *irnd++ );
    o( 0x02e9c1 );		// shr ecx, 2
    o( 0x74 );			// jz end
    g( brnd[4] + brnd[5] + brnd[6] + 38 );
    oad(0xb8, wrnd[2] );	// mov eax,
    o(0x75c085);		// test eax, eax, jnz
    g( brnd[4] + 1 );
    for( i = 0; i <= brnd[4]; i++ )
       g( *irnd++ );
    oad(0x35, wrnd[3] );	// xor eax,
    g(0x75);			// jnz
    g( brnd[5] + 1 );
    for( i = 0; i <= brnd[5]; i++ )
       g( *irnd++ );
    prolog.lfsr = wrnd[2] ^ wrnd[3];
    o( 0xe8d149 );		// dec ecx, shr eax, 1
    o( 0x0573 );		// jnc
    oad(0x35, 0xa451c123 );	// xor eax, poly
    g(0x75);			// jnz
    g( brnd[6] + 1 );
    for( i = 0; i <= brnd[6]; i++ )
       g( *irnd++ );
    o( 0x8b0431 );		// xor [ebx+ecx*4], eax
    o( 0x75c985 );		// test ecx, ecx, jnz
    g( -(brnd[6] + 20 ) );
    oad(0xbb, wrnd[5] );	// mov ebx, _start
    prolog.start_entry = ind - 4;
    o(0x75db85);		// test ebx, ebx, jnz
    g( brnd[7] + 1 );
    for( i = 0; i <= brnd[7]; i++ )
       g( *irnd++ );
    oad(0xc381, -wrnd[5] );	// add	ebx
    g(0x75);			// jnz
    g( brnd[8] + 1 );
    for( i = 0; i <= brnd[8]; i++ )
       g( *irnd++ );
    o( 0xe3ff );		// jmp ebx
}

/* generate function prolog of type 't' */
ST_FUNC void gfunc_prolog(CType *func_type)
{
    int addr, align, size, func_call, fastcall_nb_regs;
    int param_index, param_addr;
    uint8_t *fastcall_regs_ptr;
    Sym *sym;
    CType *type;

    if( xx.sw.prolog ) {
       read( rnd_fd, &prolog, 2 * sizeof( u32 ) );
       prolog.ebp_offset = prolog.add_ebp + prolog.add_esp;
    }

    sym = func_type->ref;
    func_call = sym->f.func_call;
    addr = 8 + 4;
    loc = 0;
    func_vc = 0;

    if (func_call >= FUNC_FASTCALL1 && func_call <= FUNC_FASTCALL3) {
        fastcall_nb_regs = func_call - FUNC_FASTCALL1 + 1;
        fastcall_regs_ptr = fastcall_regs;
    } else if (func_call == FUNC_FASTCALLW) {
        fastcall_nb_regs = 2;
        fastcall_regs_ptr = fastcallw_regs;
    } else {
        fastcall_nb_regs = 0;
        fastcall_regs_ptr = NULL;
    }
    param_index = 0;
#ifdef TCC_TARGET_PE
    ind += ( xx.sw.prolog ? X_PROLOG_SIZE : FUNC_PROLOG_SIZE);
#else
    ind += ( xx.sw.prolog ? X_PROLOG_SIZE : FUNC_PROLOG_SIZE);
#endif
    func_sub_sp_offset = ind;
    /* if the function returns a structure, then add an
       implicit pointer parameter */
    func_vt = sym->type;
    func_var = (sym->f.func_type == FUNC_ELLIPSIS);
#ifdef TCC_TARGET_PE
    size = type_size(&func_vt,&align);
    if (((func_vt.t & VT_BTYPE) == VT_STRUCT)
        && (size > 8 || (size & (size - 1)))) {
#else
    if ((func_vt.t & VT_BTYPE) == VT_STRUCT) {
#endif
        /* XXX: fastcall case ? */
        func_vc = addr;
        addr += 4;
        param_index++;
    }
    /* define parameters */
    while ((sym = sym->next) != NULL) {
        type = &sym->type;
        size = type_size(type, &align);
        size = (size + 3) & ~3;
#ifdef FUNC_STRUCT_PARAM_AS_PTR
        /* structs are passed as pointer */
        if ((type->t & VT_BTYPE) == VT_STRUCT) {
            size = 4;
        }
#endif
        if (param_index < fastcall_nb_regs) {
            /* save FASTCALL register */
            loc -= 4;
            o(0x89);     /* movl */
            gen_modrm(fastcall_regs_ptr[param_index], VT_LOCAL, NULL, loc);
            param_addr = loc;
        } else {
            param_addr = addr;
            addr += size;
        }
        sym_push(sym->v & ~SYM_FIELD, type,
                 VT_LOCAL | lvalue_type(type->t), param_addr);
        param_index++;
    }
    func_ret_sub = 0;
    /* pascal type call or fastcall ? */
    if (func_call == FUNC_STDCALL || func_call == FUNC_FASTCALLW)
        func_ret_sub = addr - 8;
#ifndef TCC_TARGET_PE
    else if (func_vc)
        func_ret_sub = 4;
#endif

#ifdef CONFIG_TCC_BCHECK
    /* leave some room for bound checking code */
    if (tcc_state->do_bounds_check) {
        func_bound_offset = lbounds_section->data_offset;
        func_bound_ind = ind;
        oad(0xb8, 0); /* lbound section pointer */
        oad(0xb8, 0); /* call to function */
    }
#endif
}

/* generate function epilog */
ST_FUNC void gfunc_epilog(void)
{
    addr_t v, saved_ind;
    int	i;
    u8 rnd[0x10];

#ifdef CONFIG_TCC_BCHECK
    if (tcc_state->do_bounds_check
     && func_bound_offset != lbounds_section->data_offset) {
        addr_t saved_ind;
        addr_t *bounds_ptr;
        Sym *sym_data;

        /* add end of table info */
        bounds_ptr = section_ptr_add(lbounds_section, sizeof(addr_t));
        *bounds_ptr = 0;

        /* generate bound local allocation */
        saved_ind = ind;
        ind = func_bound_ind;
        sym_data = get_sym_ref(&char_pointer_type, lbounds_section, 
                               func_bound_offset, lbounds_section->data_offset);
        greloc(cur_text_section, sym_data,
               ind + 1, R_386_32);
        oad(0xb8, 0); /* mov %eax, xxx */
        gen_static_call(TOK___bound_local_new);
        ind = saved_ind;

        /* generate bound check local freeing */
        o(0x5250); /* save returned value, if any */
        greloc(cur_text_section, sym_data, ind + 1, R_386_32);
        oad(0xb8, 0); /* mov %eax, xxx */
        gen_static_call(TOK___bound_local_delete);
        o(0x585a); /* restore returned value, if any */
    }
#endif

    /* align local size to word & save local variables */
    v = (-loc + 3) & -4;

    if( xx.sw.prolog ) {
      o(0xec89);		      /* mov esp, ebp */
      oad(0xec81, prolog.ebp_offset); /* sub esp, */
      o(0x5d);			      /* pop ebp */
    } else
    o(0xc9); /* leave */
    o(0x5b);			      /* pop ebx */
    if (func_ret_sub == 0) {
      if( xx.sw.epilog ) {
	read( rnd_fd, rnd, 0x10 );
	rnd[0] &= 0xf;
	o(0x59);	/* pop ecx */
	o(0x75);	/* jnz */
	o( rnd[0] + 1 );
	for( i = 0; i <= rnd[0]; i++ )
	  g( rnd[i] );
	o( 0xe1ff );	/* jmp ecx */
      } else
        o(0xc3); /* ret */
    } else {
        o(0xc2); /* ret n */
        g(func_ret_sub);
        g(func_ret_sub >> 8);
    }
    saved_ind = ind;
    ind = func_sub_sp_offset - ( xx.sw.prolog ? X_PROLOG_SIZE : FUNC_PROLOG_SIZE );
#ifdef TCC_TARGET_PE
    if( xx.sw.prolog )
      read( rnd_fd, rnd, 4 );
    if (v >= 4096) {
	if( xx.sw.prolog ) {
	  oad(0xbb, *(u32 *)rnd );
	  o( 0xd887);	      // xchg eax, ebx
	  oad(0xbb, *(u32 *)rnd >> 1 ^ v );
	  o( 0xe8d1);	      // shr  eax, 1
	  o( 0xd831);	      // xor  eax, ebx
	} else
        oad(0xb8, v); /* mov stacksize, %eax */
        gen_static_call(TOK___chkstk); /* call __chkstk, (does the stackframe too) */
    } else
#endif
    {
      o(0x53);				 /* push ebx */
      if( xx.sw.prolog ) {
	o(0x55);			 // push ebp
	oad(0xc481, prolog.add_esp);	 // add esp,
	o(0xe589);			 // mov ebp, esp
	oad(0xc581, prolog.add_ebp);	 // add ebp,
	oad(0xec81, prolog.add_esp + v); // sub esp, stacksize
#ifdef TCC_TARGET_PE
	oad(0xbb, *(u32 *)rnd );
	o(0x90);  // adjust to X_PROLOG_SIZE
#endif
      } else {
        o(0xe58955);  /* push %ebp, mov %esp, %ebp */
        o(0xec81);  /* sub esp, stacksize */
        gen_le32(v);
#ifdef TCC_TARGET_PE
        o(0x90);  /* adjust to FUNC_PROLOG_SIZE */
#endif
    }
    }
    ind = saved_ind;
}

/* generate a jump to a label */
ST_FUNC int gjmp(int t)
{
    int i, r;
    u8 rnd[0x58];

    if( xx.sw.jmp ) {
	read( rnd_fd, rnd, 0x58 );
	rnd[4] &= 0xf;
	rnd[5] &= 0xf;
	rnd[6] &= 0xf;
	rnd[7] &= 0xf;
	rnd[8] &= 0xf;
	oad(0xe8, rnd[4] + 1 );
	for( i = 0; i <= rnd[4]; i++ )
	  g( rnd[0x08+i] );
	o(0x5b);    // pop ebx
	o(0xdb21);  // and ebx, ebx
	o(0x75);    // jnz
	o( rnd[5] + 1 );
	for( i = 0; i <= rnd[5]; i++ )
	  g( rnd[0x18+i] );
	oad(0xc381, *(u32 *)rnd + rnd[4] + rnd[5] + rnd[6] + 22); /* add ebx, */
	o(0x75);    // jnz
	o( rnd[6] + 1 );
	for( i = 0; i <= rnd[6]; i++ )
	  g( rnd[0x28+i] );
	r = oad(0xc381, t ); // add ebx,
	o(0x75);	     // jnz
	o( rnd[7] + 1 );
	for( i = 0; i <= rnd[7]; i++ )
	  g( rnd[0x38+i] );
	oad(0xeb81, *(u32 *)rnd ); // sub ebx,
	o(0x75);		   // jnz
	o( rnd[8] + 1 );
	for( i = 0; i <= rnd[8]; i++ )
	  g( rnd[0x48+i] );
	o(0xe3ff);
	return r;
    } else
    return gjmp2(0xe9, t);
}

/* generate a jump to a fixed address */
ST_FUNC void gjmp_addr(int a)
{
    int r;
    r = a - ind - 2;
    if (r == (char)r) {
        g(0xeb);
        g(r);
    } else {
        oad(0xe9, a - ind - 5);
    }
}

ST_FUNC void gtst_addr(int inv, int a)
{
    int v = vtop->r & VT_VALMASK;
    if (v == VT_CMP) {
	inv ^= (vtop--)->c.i;
	a -= ind + 2;
	if (a == (char)a) {
	    g(inv - 32);
	    g(a);
	} else {
	    g(0x0f);
	    oad(inv - 16, a - 4);
	}
    } else if ((v & ~1) == VT_JMP) {
	if ((v & 1) != inv) {
	    gjmp_addr(a);
	    gsym(vtop->c.i);
	} else {
	    gsym(vtop->c.i);
	    o(0x05eb);
	    gjmp_addr(a);
	}
	vtop--;
    }
}

/* generate a test. set 'inv' to invert test. Stack entry is popped */
ST_FUNC int gtst(int inv, int t)
{
    int v = vtop->r & VT_VALMASK;
    if (nocode_wanted) {
        ;
    } else if (v == VT_CMP) {
        /* fast case : can jump directly since flags are set */
        g(0x0f);
        t = gjmp2((vtop->c.i - 16) ^ inv, t);
    } else if (v == VT_JMP || v == VT_JMPI) {
        /* && or || optimization */
        if ((v & 1) == inv) {
            /* insert vtop->c jump list in t */
            uint32_t n1, n = vtop->c.i;
            if (n) {
                while ((n1 = read32le(cur_text_section->data + n)))
                    n = n1;
                write32le(cur_text_section->data + n, t);
                t = vtop->c.i;
            }
        } else {
            t = gjmp(t);
            gsym(vtop->c.i);
        }
    }
    vtop--;
    return t;
}

/* generate an integer binary operation */
ST_FUNC void gen_opi(int op)
{
    int r, fr, opc, c;

    switch(op) {
    case '+':
    case TOK_ADDC1: /* add with carry generation */
        opc = 0;
    gen_op8:
        if ((vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) == VT_CONST) {
            /* constant case */
            vswap();
            r = gv(RC_INT);
            vswap();
            c = vtop->c.i;
            if (c == (char)c) {
                /* generate inc and dec for smaller code */
                if (c==1 && opc==0 && op != TOK_ADDC1) {
                    o (0x40 | r); // inc
                } else if (c==1 && opc==5 && op != TOK_SUBC1) {
                    o (0x48 | r); // dec
                } else {
                    o(0x83);
                    o(0xc0 | (opc << 3) | r);
                    g(c);
                }
            } else {
                o(0x81);
                oad(0xc0 | (opc << 3) | r, c);
            }
        } else {
            gv2(RC_INT, RC_INT);
            r = vtop[-1].r;
            fr = vtop[0].r;
            o((opc << 3) | 0x01);
            o(0xc0 + r + fr * 8); 
        }
        vtop--;
        if (op >= TOK_ULT && op <= TOK_GT) {
            vtop->r = VT_CMP;
            vtop->c.i = op;
        }
        break;
    case '-':
    case TOK_SUBC1: /* sub with carry generation */
        opc = 5;
        goto gen_op8;
    case TOK_ADDC2: /* add with carry use */
        opc = 2;
        goto gen_op8;
    case TOK_SUBC2: /* sub with carry use */
        opc = 3;
        goto gen_op8;
    case '&':
        opc = 4;
        goto gen_op8;
    case '^':
        opc = 6;
        goto gen_op8;
    case '|':
        opc = 1;
        goto gen_op8;
    case '*':
        gv2(RC_INT, RC_INT);
        r = vtop[-1].r;
        fr = vtop[0].r;
        vtop--;
        o(0xaf0f); /* imul fr, r */
        o(0xc0 + fr + r * 8);
        break;
    case TOK_SHL:
        opc = 4;
        goto gen_shift;
    case TOK_SHR:
        opc = 5;
        goto gen_shift;
    case TOK_SAR:
        opc = 7;
    gen_shift:
        opc = 0xc0 | (opc << 3);
        if ((vtop->r & (VT_VALMASK | VT_LVAL | VT_SYM)) == VT_CONST) {
            /* constant case */
            vswap();
            r = gv(RC_INT);
            vswap();
            c = vtop->c.i & 0x1f;
            o(0xc1); /* shl/shr/sar $xxx, r */
            o(opc | r);
            g(c);
        } else {
            /* we generate the shift in ecx */
            gv2(RC_INT, RC_ECX);
            r = vtop[-1].r;
            o(0xd3); /* shl/shr/sar %cl, r */
            o(opc | r);
        }
        vtop--;
        break;
    case '/':
    case TOK_UDIV:
    case TOK_PDIV:
    case '%':
    case TOK_UMOD:
    case TOK_UMULL:
        /* first operand must be in eax */
        /* XXX: need better constraint for second operand */
        gv2(RC_EAX, RC_ECX);
        r = vtop[-1].r;
        fr = vtop[0].r;
        vtop--;
        save_reg(TREG_EDX);
        /* save EAX too if used otherwise */
        save_reg_upstack(TREG_EAX, 1);
        if (op == TOK_UMULL) {
            o(0xf7); /* mul fr */
            o(0xe0 + fr);
            vtop->r2 = TREG_EDX;
            r = TREG_EAX;
        } else {
            if (op == TOK_UDIV || op == TOK_UMOD) {
                o(0xf7d231); /* xor %edx, %edx, div fr, %eax */
                o(0xf0 + fr);
            } else {
                o(0xf799); /* cltd, idiv fr, %eax */
                o(0xf8 + fr);
            }
            if (op == '%' || op == TOK_UMOD)
                r = TREG_EDX;
            else
                r = TREG_EAX;
        }
        vtop->r = r;
        break;
    default:
        opc = 7;
        goto gen_op8;
    }
}

/* generate a floating point operation 'v = t1 op t2' instruction. The
   two operands are guaranteed to have the same floating point type */
/* XXX: need to use ST1 too */
ST_FUNC void gen_opf(int op)
{
    int a, ft, fc, swapped, r;

    /* convert constants to memory references */
    if ((vtop[-1].r & (VT_VALMASK | VT_LVAL)) == VT_CONST) {
        vswap();
        gv(RC_FLOAT);
        vswap();
    }
    if ((vtop[0].r & (VT_VALMASK | VT_LVAL)) == VT_CONST)
        gv(RC_FLOAT);

    /* must put at least one value in the floating point register */
    if ((vtop[-1].r & VT_LVAL) &&
        (vtop[0].r & VT_LVAL)) {
        vswap();
        gv(RC_FLOAT);
        vswap();
    }
    swapped = 0;
    /* swap the stack if needed so that t1 is the register and t2 is
       the memory reference */
    if (vtop[-1].r & VT_LVAL) {
        vswap();
        swapped = 1;
    }
    if (op >= TOK_ULT && op <= TOK_GT) {
        /* load on stack second operand */
        load(TREG_ST0, vtop);
        save_reg(TREG_EAX); /* eax is used by FP comparison code */
        if (op == TOK_GE || op == TOK_GT)
            swapped = !swapped;
        else if (op == TOK_EQ || op == TOK_NE)
            swapped = 0;
        if (swapped)
            o(0xc9d9); /* fxch %st(1) */
        if (op == TOK_EQ || op == TOK_NE)
            o(0xe9da); /* fucompp */
        else
            o(0xd9de); /* fcompp */
        o(0xe0df); /* fnstsw %ax */
        if (op == TOK_EQ) {
            o(0x45e480); /* and $0x45, %ah */
            o(0x40fC80); /* cmp $0x40, %ah */
        } else if (op == TOK_NE) {
            o(0x45e480); /* and $0x45, %ah */
            o(0x40f480); /* xor $0x40, %ah */
            op = TOK_NE;
        } else if (op == TOK_GE || op == TOK_LE) {
            o(0x05c4f6); /* test $0x05, %ah */
            op = TOK_EQ;
        } else {
            o(0x45c4f6); /* test $0x45, %ah */
            op = TOK_EQ;
        }
        vtop--;
        vtop->r = VT_CMP;
        vtop->c.i = op;
    } else {
        /* no memory reference possible for long double operations */
        if ((vtop->type.t & VT_BTYPE) == VT_LDOUBLE) {
            load(TREG_ST0, vtop);
            swapped = !swapped;
        }
        
        switch(op) {
        default:
        case '+':
            a = 0;
            break;
        case '-':
            a = 4;
            if (swapped)
                a++;
            break;
        case '*':
            a = 1;
            break;
        case '/':
            a = 6;
            if (swapped)
                a++;
            break;
        }
        ft = vtop->type.t;
        fc = vtop->c.i;
        if ((ft & VT_BTYPE) == VT_LDOUBLE) {
            o(0xde); /* fxxxp %st, %st(1) */
            o(0xc1 + (a << 3));
        } else {
            /* if saved lvalue, then we must reload it */
            r = vtop->r;
            if ((r & VT_VALMASK) == VT_LLOCAL) {
                SValue v1;
                r = get_reg(RC_INT);
                v1.type.t = VT_INT;
                v1.r = VT_LOCAL | VT_LVAL;
                v1.c.i = fc;
                load(r, &v1);
                fc = 0;
            }

            if ((ft & VT_BTYPE) == VT_DOUBLE)
                o(0xdc);
            else
                o(0xd8);
            gen_modrm(a, r, vtop->sym, fc);
        }
        vtop--;
    }
}

/* convert integers to fp 't' type. Must handle 'int', 'unsigned int'
   and 'long long' cases. */
ST_FUNC void gen_cvt_itof(int t)
{
    save_reg(TREG_ST0);
    gv(RC_INT);
    if ((vtop->type.t & VT_BTYPE) == VT_LLONG) {
        /* signed long long to float/double/long double (unsigned case
           is handled generically) */
        o(0x50 + vtop->r2); /* push r2 */
        o(0x50 + (vtop->r & VT_VALMASK)); /* push r */
        o(0x242cdf); /* fildll (%esp) */
        o(0x08c483); /* add $8, %esp */
    } else if ((vtop->type.t & (VT_BTYPE | VT_UNSIGNED)) == 
               (VT_INT | VT_UNSIGNED)) {
        /* unsigned int to float/double/long double */
        o(0x6a); /* push $0 */
        g(0x00);
        o(0x50 + (vtop->r & VT_VALMASK)); /* push r */
        o(0x242cdf); /* fildll (%esp) */
        o(0x08c483); /* add $8, %esp */
    } else {
        /* int to float/double/long double */
        o(0x50 + (vtop->r & VT_VALMASK)); /* push r */
        o(0x2404db); /* fildl (%esp) */
        o(0x04c483); /* add $4, %esp */
    }
    vtop->r = TREG_ST0;
}

/* convert fp to int 't' type */
ST_FUNC void gen_cvt_ftoi(int t)
{
    int bt = vtop->type.t & VT_BTYPE;
    if (bt == VT_FLOAT)
        vpush_global_sym(&func_old_type, TOK___fixsfdi);
    else if (bt == VT_LDOUBLE)
        vpush_global_sym(&func_old_type, TOK___fixxfdi);
    else
        vpush_global_sym(&func_old_type, TOK___fixdfdi);
    vswap();
    gfunc_call(1);
    vpushi(0);
    vtop->r = REG_IRET;
    vtop->r2 = REG_LRET;
}

/* convert from one floating point type to another */
ST_FUNC void gen_cvt_ftof(int t)
{
    /* all we have to do on i386 is to put the float in a register */
    gv(RC_FLOAT);
}

/* computed goto support */
ST_FUNC void ggoto(void)
{
    gcall_or_jmp(1);
    vtop--;
}

/* bound check support functions */
#ifdef CONFIG_TCC_BCHECK

/* generate a bounded pointer addition */
ST_FUNC void gen_bounded_ptr_add(void)
{
    /* prepare fast i386 function call (args in eax and edx) */
    gv2(RC_EAX, RC_EDX);
    /* save all temporary registers */
    vtop -= 2;
    save_regs(0);
    /* do a fast function call */
    gen_static_call(TOK___bound_ptr_add);
    /* returned pointer is in eax */
    vtop++;
    vtop->r = TREG_EAX | VT_BOUNDED;
    /* address of bounding function call point */
    vtop->c.i = (cur_text_section->reloc->data_offset - sizeof(Elf32_Rel));
}

/* patch pointer addition in vtop so that pointer dereferencing is
   also tested */
ST_FUNC void gen_bounded_ptr_deref(void)
{
    addr_t func;
    int  size, align;
    Elf32_Rel *rel;
    Sym *sym;

    size = 0;
    /* XXX: put that code in generic part of tcc */
    if (!is_float(vtop->type.t)) {
        if (vtop->r & VT_LVAL_BYTE)
            size = 1;
        else if (vtop->r & VT_LVAL_SHORT)
            size = 2;
    }
    if (!size)
        size = type_size(&vtop->type, &align);
    switch(size) {
    case  1: func = TOK___bound_ptr_indir1; break;
    case  2: func = TOK___bound_ptr_indir2; break;
    case  4: func = TOK___bound_ptr_indir4; break;
    case  8: func = TOK___bound_ptr_indir8; break;
    case 12: func = TOK___bound_ptr_indir12; break;
    case 16: func = TOK___bound_ptr_indir16; break;
    default:
        tcc_error("unhandled size when dereferencing bounded pointer");
        func = 0;
        break;
    }

    /* patch relocation */
    /* XXX: find a better solution ? */
    rel = (Elf32_Rel *)(cur_text_section->reloc->data + vtop->c.i);
    sym = external_global_sym(func, &func_old_type, 0);
    if (!sym->c)
        put_extern_sym(sym, NULL, 0, 0);
    rel->r_info = ELF32_R_INFO(sym->c, ELF32_R_TYPE(rel->r_info));
}
#endif

/* Save the stack pointer onto the stack */
ST_FUNC void gen_vla_sp_save(int addr) {
    /* mov %esp,addr(%ebp)*/
    o(0x89);
    gen_modrm(TREG_ESP, VT_LOCAL, NULL, addr);
}

/* Restore the SP from a location on the stack */
ST_FUNC void gen_vla_sp_restore(int addr) {
    o(0x8b);
    gen_modrm(TREG_ESP, VT_LOCAL, NULL, addr);
}

/* Subtract from the stack pointer, and push the resulting value onto the stack */
ST_FUNC void gen_vla_alloc(CType *type, int align) {
#ifdef TCC_TARGET_PE
    /* alloca does more than just adjust %rsp on Windows */
    vpush_global_sym(&func_old_type, TOK_alloca);
    vswap(); /* Move alloca ref past allocation size */
    gfunc_call(1);
#else
    int r;
    r = gv(RC_INT); /* allocation size */
    /* sub r,%rsp */
    o(0x2b);
    o(0xe0 | r);
    /* We align to 16 bytes rather than align */
    /* and ~15, %esp */
    o(0xf0e483);
    vpop();
#endif
}

/* end of X86 code generator */
/*************************************************************/
#endif
/*************************************************************/
